# -*- coding: utf-8 -*-
"""
Created on Wed Feb 11 14:13:49 2015

@author: Carl-Magnus Svensson
@affiliation: Research Group Applied Systems Biology, Leibniz Institute for 
Natural Product Research and Infection Biology – Hans Knöll Institute (HKI),
Beutenbergstrasse 11a, 07745 Jena, Germany.
@email: carl-magnus.svensson@leibniz-hki.de or cmgsvensson@gmail.com

This is an implementation of the TREM algorithm for cluster segmentation.The 
full details of TREM can be found in Svensson et al. (2017), Segmentation of 
clusters by template rotation expectation maximization, CVIU, 154:64-72 on 
http://www.sciencedirect.com/science/article/pii/S1077314216301114. If any part 
of the code is used for academic purposes or publications, please cite the 
above mentioned paper.

Copyright (c) 2016-2017, Dr. Carl-Magnus Svensson

Research Group Applied Systems Biology - Head: Prof. Dr. Marc Thilo Figge
https://www.leibniz-hki.de/en/applied-systems-biology.html
HKI-Center for Systems Biology of Infection
Leibniz Institute for Natural Product Research and Infection Biology -
Hans Knöll Insitute (HKI)
Adolf-Reichwein-Straße 23, 07745 Jena, Germany

Licence: BSD-3-Clause, see ./LICENSE or 
https://opensource.org/licenses/BSD-3-Clause for full details

"""
__version__ = "0.1"
import numpy as np
import numpy.linalg as nl
from scipy.optimize import minimize

class TREM:
    def __init__(self, mu=np.atleast_2d([0,0]), Sigma0 = np.diag([1,1]), C = np.diag([1,1]), K = 1,
                 ll_tol = 0.1, mu_tol = 0.01, Sigma_tol = 0.0001, non_linear_maximization = False):
        ''' Implementation of the template rotation expectation maximization (TREM)
        
            Parameters:
            -------------------------------------------------------------------
            mu: array_type, float, Kx2x1
                Initial guess for the means of the GMM
            Sigma0: array_type, float, 2x2
                The fixed covariance matrixt that is the shape template in TREM
            C: array_type, float, Kx2x2
                Covariance matrices that is used to calculate covariance 
                matrices for the GMM by sqrt{Sigma0} P sqrt{Sigma0}
            K: int
                Number of Gaussians in the GMM
            ll_tol: float
                Tolerance value for the log-likelihood during EM. When the
                improvement is less than ll_tol the iteration stops
            mu_tol: float
                Tolerance value for the shift of gaussians during EM. When the
                total translation of the Gaussians is less than mu_tol the
                iteration stops
            Sigma_tol: float
                Tolerance value for the change of covariance matrix during EM. 
                When the change of Sigma (2-norm) is less than Sigma_tol the
                iteration stops
            non_linear_maximization: bool
                Determines if a geenral purpose should be used (GPR-EMT). If 
                false the EM based TREM will be used
                
        '''
        self.ll = 0.0
        self.Sigma0 = Sigma0
        self.Sigma0_inv = np.diag(1.0/np.diag(Sigma0))
        
        self.K = K
        self.mu = mu
        self.Sigma = self._calc_Sigma(C)
        self.prior = [1/K]*K
        
        # Iteration stop criteria
        self.ll_tol = ll_tol
        self.mu_tol = mu_tol
        self.Sigma_tol = Sigma_tol
        self.non_linear_maximization = non_linear_maximization
        
    def log_norm(self, x, mu=None, Sigma=None):
        ''' Logarithm of the normal distribution
        
            Parameters:
            -------------------------------------------------------------------
            x: array_type, float/int, 2x1
                Positions of the non-zero pixels fitted by TREM
            mu: array_type, float, 2x1
                The mean of the normal distribution
            Sigma: array_type, float; 2x2
                Covariance matrix of the normal distribution, in TREM a rotated 
                version of Sigma0
                
            Returns:
            -------------------------------------------------------------------
            log_p: float
                The logarithmic (base e) value of the PDF of the Normal 
                distribution
        '''
        if mu is None:
            mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
            
        return -np.log(2*np.pi) - 0.5*np.log(nl.det(Sigma)) - 0.5*np.inner((x-mu),np.inner(nl.inv(Sigma),x-mu))
        
    def q_G_x(self, x, mu=None, Sigma=None, prior=None):
        ''' 
            PDF q that is set in the E-step and is then used during the M-step
            of the EM iteration
        
            Parameters:
            -----------------------------------------------------------
            x: array_type, float/int, 2x1
                Positions of the non-zero pixels fitted by TREM
            mu: array_type, float, 2x1
                The mean of the normal distribution
            Sigma: array_type, float; 2x2
                Covariance matrix of the normal distribution, in TREM a rotated 
                version of Sigma0
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K
                
            Returns:
            -------------------------------------------------------------------
            q: float
                Value of the PDF
                
        '''
        if mu is None:
            mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
        if prior is None:
            prior = self.prior[:]
        try:
            return np.exp(self.log_norm(x, mu, Sigma))*prior
        except:
            return 0.0
                
        
    def log_likelihood(self, data, mu=None, Sigma=None, prior = [1.0]):
        ''' The log likelihood of the GMM given the data
        
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            Sigma: array_type, float; Kx2x2
                Covariance matrix of the normal distributions, in TREM a rotated 
                version of Sigma0
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K
                
            Returns:
            -------------------------------------------------------------------
            ll: float
                The likelyhood of the data given the model defined by *mu*, 
                *Sigma* and *prior*
                
        '''
        self.ll = 0.0
        if mu is None:
            mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
        if prior is None:
            prior = self.prior[:]
        for d in data:
            l_k = 0.0
            for k in xrange(self.K):
                l_k += np.exp(self.log_norm(d, mu[k].T, Sigma[k])+0.5)*prior[k]
            self.ll += np.log(l_k)
            
        return self.ll 
        
    def mu_update(self, data, old_mu=None, Sigma=None, Beta = 1.0, prior = None):
        ''' 
            Standard update rule for means of the normal distributions in a 
            GMM
        
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            old_mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            Sigma: array_type, float; Kx2x2
                Covariance matrix of the normal distributions, in TREM a rotated 
                version of Sigma0
            Beta: float
                Parameter controlling the deterministic annealing
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K
                
            Returns:
            -------------------------------------------------------------------
            mu_out: array_type, float, Kx2
                The updated mean positions of the Gaussians
                
        '''
        if old_mu is None:
            old_mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
        if prior is None:
            prior = self.prior[:]
        old_mu = np.array(old_mu)
        mu_out = np.zeros(old_mu.shape)
                
        for k in xrange(self.K):
            N_k = 0.0
            for d in data:
                q = self.q_G_x(d, old_mu[k], Sigma[k], prior[k])
                q_denom = q
                for k_prime in xrange(self.K):
                    if k_prime != k:
                        q_denom += self.q_G_x(d, old_mu[k_prime], Sigma[k_prime], prior[k_prime])
                mu_out[k] += (q/q_denom)**Beta*d
                N_k += (q/q_denom)**Beta
            mu_out[k] /= N_k
            
        return mu_out   
        
    def prior_update(self, data, mu = None, Sigma = None, Beta = 1.0, prior = None):
        ''' 
            Standard update rule for priors in a GMM
        
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            old_mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            Sigma: array_type, float; Kx2x2
                Covariance matrix of the normal distributions, in TREM a rotated 
                version of Sigma0
            Beta: float
                Parameter controlling the deterministic annealing
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K

            Returns:
            -------------------------------------------------------------------
            p_out: array_type, float, K
                The updated prior                
        '''
                
        if mu is None:
            mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
        if prior is None:
            prior = self.prior[:]
            
        p_out = np.zeros(self.K)
        N = data.shape[0]
        for k in xrange(self.K):
            N_k = 0.0
            for d in data:
                q = self.q_G_x(d, mu[k], Sigma[k], prior[k])
                q_denom = q
                for k_prime in xrange(self.K):
                    if k_prime != k:
                        q_denom += self.q_G_x(d, mu[k_prime], Sigma[k_prime], prior[k_prime])
                N_k += (q/q_denom)**Beta
            p_out[k] = 1.0/N*N_k
            
        return p_out        
        
    def C_update(self, data, mu = None, Sigma = None, Beta = 1.0, prior = None):
        ''' Update rule for means of the covariance matrix used to calculate the
            covariance matrix for the normal distributions in a GMM by
            sqrt{Sigma0} P sqrt{Sigma0}
        
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            Sigma: array_type, float; Kx2x2
                Covariance matrix of the normal distributions, in TREM a rotated 
                version of Sigma0
            Beta: float
                Parameter controlling the deterministic annealing
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K

            Returns:
            -------------------------------------------------------------------
            C_out: array_type, float; Kx2x2
                The updated scaling matrix
        '''
        if mu is None:
            mu = self.mu[:]
        if Sigma is None:
            Sigma = self.Sigma[:]
        if prior is None:
            prior = self.prior[:]
            
        C_out = np.zeros([self.K,2,2])
        for k in xrange(self.K):
            N_k = 0.0
            for d in data:
                #Q[k,ii] = self.q_G_x(data[ii], mu[k], Sigma[k], prior[k])
                q = self.q_G_x(d, mu[k], Sigma[k], prior[k])
                d_x = np.atleast_2d(d-mu[k]).T
                q_denom = q
                for k_prime in xrange(self.K):
                    if k_prime != k:
                        q_denom += self.q_G_x(d, mu[k_prime], Sigma[k_prime], prior[k_prime])
                C_out[k] += (q/q_denom)**Beta*np.dot( np.dot(np.sqrt(self.Sigma0_inv), d_x), np.dot(d_x.T, np.sqrt(self.Sigma0_inv)) )
                
                N_k += (q/q_denom)**Beta

            C_out[k] /= (np.sqrt(nl.det(C_out[k])))
            
        return C_out
    
    def find_phi(self, phi, C):
        
        out = 0.0
        for k in xrange(self.K):
            R = np.array([[np.cos(phi[k]), -np.sin(phi[k])],[np.sin(phi[k]), np.cos(phi[k])]])
            out += nl.norm(np.dot(np.sqrt(self.Sigma0_inv),np.dot(R,np.dot(self.Sigma0,np.dot(R.T,np.sqrt(self.Sigma0_inv))))) - C[k])
        
        return out
        
    def rotate_Sigma0(self, phi): 
        ''' Calculate the covariance matrix by rotating it by *phi*
        
            Parameters:
            -----------------------------------------------------------
            phi: float
                Angle of rotation 
             
            Returns:
            -------------------------------------------------------------------
            Sig: array_type, float; Kx2x2
                The rotated covariance matrix   
        '''
        R = np.array([[np.cos(phi), -np.sin(phi)],[np.sin(phi), np.cos(phi)]])  
        Sig = np.dot(R, np.dot(self.Sigma0,R.T))
     
        return Sig 
        
    def log_likelihood_phi(self, phi, data, mu = None, prior = None):
        ''' Log likelihood as calculated by phi in GPR-EMT
        
            Parameters:
            -----------------------------------------------------------
            phi: float
                Angle of rotation 
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K
                
            Returns:
            -------------------------------------------------------------------
            ll: float
                The likelihood of the *data* given the model defined by *phi*, 
                *mu* and *prior*                
        '''
                
        if mu is None:
            mu = self.mu[:]
        if prior is None:
            prior = self.prior[:]
        ll = 0.0
        for d in data:
            l_k = 0.0
            for k in xrange(self.K):
                Sig = self.rotate_Sigma0(phi[k])
                
                l_k += np.exp(self.log_norm(d, mu[k], Sig)+0.5)*prior[k]
            ll += np.log(l_k)
        
        return -ll  
        
    def var_energy_phi(self, phi, data, mu = None, prior = None):
        '''
            The variational free energy of the model given *data*
            
            Parameters:
            -----------------------------------------------------------
            phi: float
                Angle of rotation 
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels in the cluster fitted by 
                TREM
            mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K
                
            Returns:
            -------------------------------------------------------------------
            F: float
               The variational free energy of the *data* given the model 
               defined by *phi*, *mu* and *prior*     
        '''
        if mu is None:
            mu = self.mu[:]
        if prior is None:
            prior = self.prior[:]
        F = 0.0
        Sig = self.rotate_Sigma0(phi[0])
        for d in data:
            F += self.q_G_x(d, mu, Sig, prior)*self.log_norm(d, mu, Sig)
        
        return -F
        
    def _calc_Sigma(self, C):
        ''' Calculate the covariance matrix by sqrt{Sigma0} C sqrt{Sigma0}.
        
            Parameters:
            -----------------------------------------------------------
            C: array_type, float; Kx2x2
                Scaling matrices in TREM.
            
            Returns:
            -------------------------------------------------------------------
            Sigma: array_type, float; Kx2x2
                Covariance matrices
            '''
        Sigma = np.zeros([self.K,2,2])
        for k in xrange(self.K):
            Sigma[k] = np.dot(np.sqrt(self.Sigma0), np.dot(C[k], np.sqrt(self.Sigma0)))
            
        return Sigma

    def predict(self, data, prior = None):
        ''' Predict the class of data points from the current GMM.
        
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels to be predicted.
            prior: array_type, float, K
                Prior distribution of the pixels , 1/K.
            
            Returns:
            ------------------------------------------------------------------- 
            p: array_type, int, N
                An array of length N that gives the class of each pixel in 
                *data*.
        '''    
        if prior is None:
            prior = self.prior[:]
        p = np.zeros(data.shape[0], dtype = np.int)
        for (d, ii) in zip(data, xrange(data.shape[0])):
            best_p = -np.inf
            p0 = 0
            for k in xrange(self.K):
                p_tmp = self.log_norm(d, self.mu[k], self.Sigma[k])*prior[k]
                if p_tmp > best_p:
                    best_p = p_tmp
                    p0 = k+1
            p[ii] = p0
            
        return p

    def run_update(self, data, C, mu = None, Beta = 1.0): 
        '''
            Parameters:
            -----------------------------------------------------------
            data: array_type, float/int, Nx2
                All positions of the non-zero pixels to be predicted.
            C: array_type, float; Kx2x2
                Scaling matrices in TREM.
            mu: array_type, float, Kx2
                The mean of the normal distributions, each row is a mean for
                one of the clusters
                
            Returns:
            ------------------------------------------------------------------- 
            mu: array_type, int/float, Kx2
                The means of the Gaussians after convergence.
            Sigma: array_type, float; Kx2x2
                The covariance matrices after convergence
            
        '''
        if mu is None:
            mu = self.mu
        prior = [1.0/self.K]*self.K
        Sigma = self._calc_Sigma(C)
        ll_old = np.inf
        mu_old = mu[:]
        Sigma_old = Sigma[:]
        ll = 0.0
        phi = np.random.uniform(0, np.pi, size = self.K)
        for k in xrange(self.K):
            Sigma[k] = self.rotate_Sigma0(phi[k])
        iterations = 0
        converged = False
        while not converged: 
            iterations += 1

            mu =  self.mu_update(data = data, old_mu = mu, Sigma = Sigma, Beta = Beta, prior = prior)
            if self.non_linear_maximization:
                # Run GPR-EMT
                for k in xrange(self.K):  
                    res = minimize(self.var_energy_phi, phi[k], args = (data,mu[k],prior[k]), tol = 0.00001, method = 'L-BFGS-B', bounds = [(-np.pi/2,np.pi/2)])
                    phi[k] = res['x']
                for k in xrange(self.K):
                    Sigma[k] = self.rotate_Sigma0(phi[k])
            else:
                # Run EM update for C to perform TREM
                C = self.C_update( data = data, mu = mu, Sigma = Sigma, Beta = Beta, prior = prior )
                Sigma = self._calc_Sigma(C)
            
            ll_old = np.copy(ll)
            ll = self.log_likelihood( data, mu, Sigma, prior )
            dist_moved = np.sqrt(np.sum((mu - mu_old)**2))
            Sigma_changed = np.sqrt(np.sum((Sigma - Sigma_old)**2))
            
            mu_old = mu[:]
            Sigma_old = Sigma[:]
            if (abs(ll_old - ll) < self.ll_tol or iterations > 1000 or (dist_moved < self.mu_tol/Beta and 
                Sigma_changed < self.Sigma_tol)):
                converged = True

        self.ll = self.log_likelihood( data, mu, Sigma, prior )             
        self.mu = mu
        self.Sigma = Sigma
        self.prior = prior
        
        return mu, Sigma
