# -*- coding: utf-8 -*-
"""
Created on Fri Mar 10 10:09:46 2017

@author: Carl-Magnus Svensson
@affiliation: Research Group Applied Systems Biology, Leibniz Institute for 
Natural Product Research and Infection Biology – Hans Knöll Institute (HKI),
Beutenbergstrasse 11a, 07745 Jena, Germany.
@email: carl-magnus.svensson@leibniz-hki.de or cmgsvensson@gmail.com

This is a script for testing the TREM algorithm for cluster segmentation.The 
full details of TREM can be found in Svensson et al. (2017), Segmentation of 
clusters by template rotation expectation maximization, CVIU, 154:64-72 on 
http://www.sciencedirect.com/science/article/pii/S1077314216301114. If any part 
of the code is used for academic purposes or publications, please cite the 
above mentioned paper.

Copyright (c) 2016-2017, Dr. Carl-Magnus Svensson

Research Group Applied Systems Biology - Head: Prof. Dr. Marc Thilo Figge
https://www.leibniz-hki.de/en/applied-systems-biology.html
HKI-Center for Systems Biology of Infection
Leibniz Institute for Natural Product Research and Infection Biology -
Hans Knöll Insitute (HKI)
Adolf-Reichwein-Straße 23, 07745 Jena, Germany

Licence: BSD-3-Clause, see ./LICENSE or 
https://opensource.org/licenses/BSD-3-Clause for full details

"""

import numpy as np
import pylab as pl
import TREM

# Load data
image = np.load('./data/canvas_pair_r6_10.npy')
# Binarize the image with a bit of noise
binary = (image+np.random.normal(scale=0.025,size=image.shape)) > 0.95

# plot data
pl.figure(1)
pl.imshow(image)
pl.figure(2)
pl.imshow(binary)

# Sigma0 is the template 
Sigma0 = np.diag([25,1])

# Extract the foreground pixels
II = np.array((binary).nonzero()).T

# Here we know that we have two objects, normally a challenging task
K = 2

C = np.zeros([K,2,2])
mu_init = np.zeros([K,2])

# set initial scaling matrices 
for k in xrange(K):
    C[k] = np.diag([1,1])    
    
# Pick a bunch of initial conditions
init_max = 6
likelihoods = np.zeros(init_max)
traces = np.zeros(init_max)
Ts = [None]*init_max
best_ll = -np.inf
for init in xrange(init_max):
    # set inital centres, here easy but normally challenging
    x1 = 20 + np.random.normal(scale = 2.5)
    y1 = 15 + np.random.normal(scale = 2.5)
    x2 = 20 + np.random.normal(scale = 2.5)
    y2 = 30 + np.random.normal(scale = 2.5)
    mu_init[0,:] = [x1,y1]
    mu_init[1,:] = [x2,y2]

    pl.plot(x1,y1, 'kx', markersize =18)
    pl.plot(x2,y2, 'gx', markersize =18)
    pl.axis('tight')

    # Initialize and run TREM
    Ts[init] = TREM.TREM(mu = mu_init,  Sigma0 = Sigma0, C = C, K = K,
               mu_tol = 0.000001, Sigma_tol = 0.000001, non_linear_maximization = False)
    m, c = Ts[init].run_update(II, C = C,  Beta = 1.0)
    
    likelihoods[init] = Ts[init].ll
    tr = 0.0
    for k in xrange(K):
        tr += np.trace(c[k])
    traces[init] = tr/K
print 'Likelihoods: %s'%likelihoods
print 'Traces: %s'%traces

# Pick the solution that has deformed the template the least
if (np.max(traces) - np.min(traces)) > 1.5:
    best_init = np.argmin(abs(traces-np.sum(Sigma0)))
else: 
    # If all traces are almost equal we pick the solution with highest likelihood
    best_init = np.argmax(likelihoods)

for k in xrange(K):
    print '-'*24
    print 'Mean %s: %s'%(k+1, Ts[best_init].mu[k])
    print 'Covariance matrix %s:\n %s'%(k+1, Ts[best_init].Sigma[k])
    print '-'*24
    
# Predict forground pixels. Each pixel gets assigned to one of the K classes
preds = Ts[best_init].predict(II)
labelled = binary[:].astype(np.uint8)
new_idx = np.max(labelled)+1
for p, jj in zip(preds, xrange(preds.shape[0])):
    if p > 1:
        labelled[II[jj,0],II[jj,1]] = p
  
# Plot the segementation     
pl.figure(3)
pl.imshow(labelled)
pl.show()